# stdlib
import os
import types
from inspect import isclass

# stdlib, in support of the the 'probe' method
import socket
import datetime
import time
import sys

# 3rd-party packages
from lxml import etree
from ncclient import manager as netconf_ssh
import paramiko
import jinja2

# local modules
from .rpcmeta import _RpcMetaExec
from .exception import RpcError
from .cfg import Resource 
from .facts import *
from . import jxml as JXML

_MODULEPATH = os.path.dirname(__file__)

class _MyTemplateLoader(jinja2.BaseLoader):
  """
  Create a jinja2 template loader class that can be used to 
  load templates from all over the filesystem, but defaults
  to the CWD and the 'templates' directory of the module
  """
  def __init__(self):
    self.paths = ['.', os.path.join(_MODULEPATH,'templates')]

  def get_source(self, environment, template):

    def _in_path(dir): 
      return os.path.exists(os.path.join(dir,template))

    path = filter(_in_path, self.paths)
    if not path:
      raise jinja2.TemplateNotFound(template)

    path = os.path.join(path[0],template)
    mtime = os.path.getmtime(path)
    with file(path) as f:
      source = f.read().decode('utf-8')
    return source, path, lambda: mtime == os.path.getmtime(path)    

_Jinja2ldr = jinja2.Environment(loader=_MyTemplateLoader())

class Device(object):

  ##### -------------------------------------------------------------------------
  ##### PROPERTIES
  ##### -------------------------------------------------------------------------

  ### ---------------------------------------------------------------------------
  ### property: hostname
  ### ---------------------------------------------------------------------------

  @property
  def hostname(self):
    """
    The hostname/ip-addr of the Junos device
    """
    return self._hostname if (self._hostname != 'localhost') else self.facts.get('hostname')

  ### ---------------------------------------------------------------------------
  ### property: user
  ### ---------------------------------------------------------------------------
  
  @property
  def user(self):
    """
    The login user accessing the Junos device
    """
    return self._auth_user

  ### ---------------------------------------------------------------------------
  ### property: password
  ### ---------------------------------------------------------------------------

  @property
  def password(self):
    """
      The login password to access the Junos deviec
    """
    return None  # read-only      

  @password.setter
  def password(self, value):
    self._password = value
  
  ### ---------------------------------------------------------------------------
  ### property: logfile
  ### ---------------------------------------------------------------------------

  @property
  def logfile(self):
    """
      simply returns the log file object
    """
    return self._logfile

  @logfile.setter
  def logfile(self, value):
    """
      assigns an opened file object to the device for logging
      If there is an open logfile, and 'value' is None/False
      then close the existing file
    """
    # got an existing file that we need to close
    if (not value) and (None != self._logfile):
      rc = self._logfile.close()
      self._logfile = False
      return rc

    if not isinstance(value, file):
      raise ValueError("value must be a file object")  

    self._logfile = value
    return self._logfile

  ### ---------------------------------------------------------------------------
  ### property: timeout
  ### ---------------------------------------------------------------------------    

  @property
  def timeout(self):
    """ the netconf timeout value """
    return self._conn.timeout

  @timeout.setter
  def timeout(self, value):
    """ set the netconf timeout value """
    self._conn.timeout = value
  
  ### ---------------------------------------------------------------------------
  ### property: facts
  ### ---------------------------------------------------------------------------

  @property
  def facts(self):
    return self._facts

  @facts.setter
  def facts(self, value):
    raise RuntimeError("facts is read-only!")

  ### ---------------------------------------------------------------------------
  ### property: manages
  ### ---------------------------------------------------------------------------

  @property
  def manages(self):
    """
    returns a list of Resource Managers/Utilities attached to this isinstance
    using the :bind(): method
    """
    return self._manages

  ##### -----------------------------------------------------------------------
  ##### OVERLOADS
  ##### -----------------------------------------------------------------------

  def __repr__(self):
    return "Device(%s)" % self.hostname

  ##### -----------------------------------------------------------------------
  ##### CONSTRUCTOR
  ##### -----------------------------------------------------------------------

  def _sshconf_lkup(self):
    sshconf_path = os.path.join(os.getenv('HOME'),'.ssh/config')
    if not os.path.exists(sshconf_path): return

    sshconf = paramiko.SSHConfig()
    sshconf.parse(open(sshconf_path,'r'))
    found = sshconf.lookup(self._hostname)
    self._hostname = found.get('hostname', self._hostname)
    self._port = found.get('port', self._port)


  def __init__(self, *vargs, **kvargs):
    """
    vargs[0] -- ALTERNATIVE for kvargs['host']

    kvargs['host'] -- REQUIRED
      device hostname or ipaddress

    kvargs['user'] -- OPTIONAL
      login user-name, uses $USER if not provided

    kvargs['password'] -- OPTIONAL
      login password.  if not provided, assumed ssh-keys are enforced

    kvargs['port'] -- OPTIONAL
      device login port (defaults to 830)
    """

    # private attributes
    try:
      self._hostname = vargs[0] if len(vargs) else kvargs['host']
    except:
      raise ValueError("You must provide the 'host' value")      
    self._port = kvargs.get('port', 830)

    self._sshconf_lkup()

    self._auth_user = kvargs['user'] if 'user' in kvargs else os.getenv('USER')
    self._auth_password = kvargs.get('password')
    self._conn = None
    self._j2ldr = _Jinja2ldr
    self._manages = []
    self._facts = {}

    # public attributes

    self.connected = False
    self.rpc = _RpcMetaExec( self )

  ##### -----------------------------------------------------------------------
  ##### Basic device methods
  ##### -----------------------------------------------------------------------

  def open( self, *vargs, **kvargs ):
    """
    opens a connection to the device using existing login/auth 
    information.  No additional options are supported; at this time
    """
    # open connection using ncclient transport
    self._conn =  netconf_ssh.connect( host=self.hostname, port=self._port,
      username=self._auth_user, password=self._auth_password,
      hostkey_verify=False )

    self.connected = True
    self.facts_refresh()
    return self

  def close( self ):
    """
    closes the connection to the device
    """
    self._conn.close_session()
    self.connected = False

  def execute( self, rpc_cmd, **kvargs ):
    """
    Executes an XML RPC and returns results as either XML or native python

    rpc_cmd 
      can either be an XML Element or xml-as-string.  In either case
      the command starts with the specific command element, i.e., not the
      <rpc> element itself

    kvargs['to_py']
      is a caller provided function that takes the response and
      will convert the results to native python types.  all kvargs
      will be passed to this function as well in the form:
      :to_py:( self, rpc_rsp, **kvargs )
    """

    if isinstance(rpc_cmd, str):
      rpc_cmd_e = etree.XML( rpc_cmd )
    elif isinstance(rpc_cmd, etree._Element):
      rpc_cmd_e = rpc_cmd
    else:
      raise ValueError("Dont know what to do with rpc of type %s" % rpc_cmd.__class__.__name__)

    # invoking a bad RPC will cause a connection object exception
    # will will be raised directly to the caller ... for now ...
    # @@@ need to trap this and re-raise accordingly.

    try:
      rpc_rsp_e = self._conn.rpc( rpc_cmd_e )._NCElement__doc
    except Exception as err:
      # err is an NCError from ncclient
      rsp = JXML.remove_namespaces( err.xml )
      raise RpcError(cmd=rpc_cmd_e, rsp=rsp)

    # for RPCs that have embedded rpc-errors, need to check for those now

    rpc_errs = rpc_rsp_e.xpath('.//rpc-error')
    if len(rpc_errs):
      raise RpcError( rpc_cmd_e, rpc_rsp_e, rpc_errs )

    # skip the <rpc-reply> element and pass the caller the first child element
    # generally speaking this is what they really want.  if they want to uplevel 
    # they can always call the getparent() method on it.

    try:
      ret_rpc_rsp = rpc_rsp_e[0]    
    except IndexError:
      # no children, so assume it means we are OK
      return True

    # if the caller provided a "to Python" conversion function, then invoke
    # that now and return the results of that function.  otherwise just return
    # the RPC results as XML

    if kvargs.get('to_py'):
      return kvargs['to_py']( self, ret_rpc_rsp, **kvargs )
    else:
      return ret_rpc_rsp

  ### ---------------------------------------------------------------------------
  ### cli - for cheating commands :-)
  ### ---------------------------------------------------------------------------

  def cli(self, command, format='text'):
    """
    Executes the CLI command and returns the CLI text output by default.

    command
      The CLI command to execute, e.g. "show version"

    format
      The return format, by default is text.  You can optionally select
      'xml' to return the XML structure.

    Notes:
      You can also use this method to obtain the XML RPC command for a given
      CLI command by using the pipe filter "| display xml rpc".  When you do
      this, the return value is the XML RPC command.  For example if you
      provide as the command "show version | display xml rpc", you will get
      back the XML Element <get-software-information>
    """
    try:
      rsp = self.rpc.cli( command, format)
      if rsp.tag == 'output': return rsp.text
      if rsp.tag == 'configuration-information': return rsp.findtext('configuration-output')
      if rsp.tag == 'rpc': return rsp[0]
      return rsp
    except:
      return "invalid command: "+command

  ### ---------------------------------------------------------------------------
  ### Template: retrieves a Jinja2 template
  ### ---------------------------------------------------------------------------

  def Template( self, filename, parent=None, gvars=None ):

    # templates are XML files, and the assumption here is that they will
    # have .xml extensions.  if the caller doesn't include any extension
    # be kind and add '.xml' for them

    if os.path.splitext(filename)[1] == '':
      filename = filename + '.xml'

    return self._j2ldr.get_template( filename, parent, gvars )

  ### ---------------------------------------------------------------------------
  ### dealing with bind aspects
  ### ---------------------------------------------------------------------------

  def bind(self, *vargs, **kvargs):
    """
    Used to attach things to this Device instance

    vargs
      a list of functions that will get bound as instance methods to 
      this Device instance

    kvargs
      name/class pairs that will create resource-managers bound as 
      instance attributes to this Device instance
    """
    if len(vargs):
      for fn in vargs:
        # check for name clashes before binding
        if hasattr(self, fn.__name__):
          raise ValueError("request attribute name %s already exists" % fn.__name__)
      for fn in vargs:
        # bind as instance method, majik.
        self.__dict__[fn.__name__] = types.MethodType(fn,self,self.__class__)
      return

    # first verify that the names do not conflict with
    # existing object attribute names

    for name in kvargs.keys():
      # check for name-clashes before binding
      if hasattr(self, name):
        raise ValueError("requested attribute name %s already exists" % name)

    # now instantiate items and bind to this :Device:
    for name,thing in kvargs.items():
      new_inst = thing(self)
      self.__dict__[name] = new_inst
      self._manages.append( name )

  ### ---------------------------------------------------------------------------
  ### facts
  ### ---------------------------------------------------------------------------    

  def facts_refresh(self):
    """
    reload the facts from the Junos device into :facts: property
    """
    for gather in FACT_LIST:
      gather(self, self._facts)
      
  ### ---------------------------------------------------------------------------
  ### probe
  ### ---------------------------------------------------------------------------  

  def probe(self, timeout=5, intvtimeout=1):
    """ 
    probe the device to determine if the Device can accept a remote connection.
    this method is meant to be called *prior* to :open():
    this method will not work with ssh-jumphost enviornments.

    :timeout: 
      the probe will report True/False if the device report connectivity
      within this timeout (secodns)

    :intvtimeout: 
      timeout interval on the socket connection.  generally you should not change
      this value, but you can if you want to twiddle the frequency of the socket
      attempts on the connection
    """
    start = datetime.datetime.now()
    end = start + datetime.timedelta(seconds=timeout)
    probe_ok = True

    while datetime.datetime.now() < end:
      s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
      s.settimeout(intvtimeout)
      try:
          s.connect( (self.hostname, self._port) )
          s.shutdown(socket.SHUT_RDWR)
          s.close()
          break
      except:
          time.sleep(1)
          pass
    else:
      elapsed = datetime.datetime.now() - start
      probe_ok = False

    return probe_ok



# 3rd-party modules
from lxml.builder import E 

# module packages
from ... import jxml as JXML
from .. import Resource

class SharedAddrBookAddr( Resource ):
  """
  [edit security address-book <ab_name> address <name>]

  Resource name: str
    <name> is the address item name

  Managed by: SharedAddrBook
    <ab_name> is the address book name, taken from parent resource    
  """
  PROPERTIES = [
    'description',
    'ip_prefix',
  ]

  def _xml_at_top(self):
    xml = self.P._xml_at_top()
    xml.find('.//address-book').append(E.address(self._name))
    return xml

  ##### -----------------------------------------------------------------------
  ##### XML reading
  ##### -----------------------------------------------------------------------

  def _xml_at_res(self, xml):
    return xml.find('.//address-book/address')

  def _xml_to_py(self, as_xml, to_py ):
    Resource._r_has_xml_status( as_xml, to_py )
    Resource.copyifexists( as_xml, 'description', to_py )
    to_py['ip_prefix'] = as_xml.find('ip-prefix').text

  ##### -----------------------------------------------------------------------
  ##### XML writing
  ##### -----------------------------------------------------------------------

  def _xml_change_ip_prefix(self, xml):
    xml.append(E('ip-prefix', self.should['ip_prefix']))
    return True

  ##### -----------------------------------------------------------------------
  ##### Manager List, Catalog
  ##### -----------------------------------------------------------------------

  def _r_list(self):
    # The parent keeps a property on this list, so just use it, yo!
    self._rlist = self.P['$addrs']

  def _r_catalog(self):
    get = self.P._xml_at_top()
    get.find('.//address-book').append(E('address'))
    got = self.D.rpc.get_config( get )
    for addr in got.xpath('.//address-book/address'):
      name = addr.find('name').text
      self._rcatalog[name] = {}
      self._xml_to_py( addr, self._rcatalog[name] )



import warnings
from contextlib import contextmanager
from copy import deepcopy
from lxml import etree

from .viewfields import ViewFields

class View(object):
  """
  View is the base-class that makes extracting values from XML 
  data appear as objects with attributes.
  """

  ITEM_NAME_XPATH = 'name'
  FIELDS = {}
  GROUPS = None

  ### -------------------------------------------------------------------------
  ### CONSTRUCTOR
  ### -------------------------------------------------------------------------

  def __init__(self, table, view_xml):
    """
    :table:
      instance of the RunstatTable

    :view_xml:
      this should be an lxml etree Elemenet object.  This 
      constructor also accepts a list with a single item/XML
    """
    # if as_xml is passed as a list, make sure it only has
    # a single item, common response from an xpath search
    if isinstance(view_xml,list):
      if 1 == len(view_xml):
        view_xml = view_xml[0]
      else:
        raise ValueError("constructor only accepts a single item")

    # now ensure that the thing provided is an lxml etree Element
    if not isinstance(view_xml,etree._Element):
      raise ValueError("constructor only accecpts lxml.etree._Element")  

    self._table = table
    self.ITEM_NAME_XPATH = table.ITEM_NAME_XPATH
    self._init_xml( view_xml )

  def _init_xml(self, given_xml):
    self._xml = given_xml
    if self.GROUPS is not None:
      self._groups = {}
      for xg_name,xg_xpath in self.GROUPS.items():
        xg_xml = self._xml.xpath(xg_xpath)
        if not len(xg_xml):  # @@@ this is technically an error; need to trap it
          continue
        self._groups[xg_name] = xg_xml[0]

  ### -------------------------------------------------------------------------
  ### PROPERTIES
  ### -------------------------------------------------------------------------

  @property 
  def T(self):
    """ return the Table instance for the View """
    return self._table

  @property
  def D(self):
    """ return the Device instance for this View """
    return self.T.D 
    
  @property 
  def name(self):
    """ return the name of view item """
    if self.ITEM_NAME_XPATH is None: return self._table.D.hostname
    if isinstance(self.ITEM_NAME_XPATH,str):
      # simple key
      return self._xml.findtext(self.ITEM_NAME_XPATH).strip()
    else:
      # composite key
#      return tuple([self.xml.findtext(i).strip() for i in self.ITEM_NAME_XPATH])
      return tuple([self.xml.xpath(i)[0].text.strip() for i in self.ITEM_NAME_XPATH])

  # ALIAS key <=> name
  key = name

  @property
  def xml(self):
    """ returns the XML associated to the item """
    return self._xml
   
  ### -------------------------------------------------------------------------
  ### METHODS
  ### -------------------------------------------------------------------------

  def keys(self):
    """ list of view keys, i.e. field names """
    return self.FIELDS.keys()

  def values(self):
    """ list of view values """
    return [getattr(self,field) for field in self.keys()]

  def items(self):
    """ list of tuple(key,value) """
    return zip(self.keys(), self.values())

  def _updater_instance(self, more):
    """ called from extend """
    if hasattr(more,'fields'):
      self.FIELDS = deepcopy(self.__class__.FIELDS)
      self.FIELDS.update(more.fields.end)

    if hasattr(more,'groups'):
      self.GROUPS = deepcopy(self.__class__.GROUPS)
      self.GROUPS.update(more.groups)

  def _updater_class(self,more):
    """ called from extend """
    if hasattr(more,'fields'):
      self.FIELDS.update(more.fields.end)

    if hasattr(more,'groups'):
      self.GROUPS.update(more.groups)

  @contextmanager
  def updater(self, fields=True, groups=False, all=True, **kvargs):
    """ 
    provide the ability for subclassing objects to extend the
    definitions of the fields.  this is implemented as a 
    context manager with the form called from the subclass 
    constructor:

      with self.extend() as more:
        more.fields = <dict>
        more.groups = <dict>   # optional
    """
    # -------------------------------------------------------------------------    
    # create a new object class so we can attach stuff to it arbitrarily.  
    # then pass that object to the caller, yo!
    # -------------------------------------------------------------------------    

    more = type('RunstatViewMore',(object,),{})()
    if fields is True:
      more.fields = RunstatMakerViewFields()

    # -------------------------------------------------------------------------
    # callback through context manager
    # -------------------------------------------------------------------------

    yield more
    updater = self._updater_class if all is True else self._updater_instance
    updater(more)

  def asview(self, view_cls ):
    """ create a new View object for this item """
    return view_cls(self._table, self._xml)
    
  def refresh(self):
    """ 
    ~~~ EXPERIMENTAL ~~~
    refresh the data from the Junos device.  this only works if the table
    provides an "args_key", does not update the original table, just this
    specific view/item
    """
    warnings.warn("Experimental method: refresh")
    
    if self._table.can_refresh is not True:
      raise RuntimeError("table does not support this feature")

    # create a new table instance that gets only the specific named
    # value of this view

    tbl_xml = self._table._rpc_get(self.name)
    new_xml = tbl_xml.xpath(self._table.ITEM_XPATH)[0]
    self._init_xml(new_xml)
    return self

  ### -------------------------------------------------------------------------
  ### OVERLOADS
  ### -------------------------------------------------------------------------

  def __repr__(self):
    """ returns the name of the View with the associate item name """
    return "%s:%s" % (self.__class__.__name__, self.name)

  def __getattr__(self,name):
    """ 
    returns a view item value, called as :obj.name:
    """
    item = self.FIELDS.get(name)
    if item is None:
      raise ValueError("Unknown field: '%s'" % name)

    if item.has_key('table'):
      # if this is a sub-table, then return that now
      return item['table'](self.D, self._xml)

    # otherwise, not a sub-table, and handle the field
    astype = item.get('astype',str)
    if item.has_key('group'):
      found = self._groups[item['group']].xpath(item['xpath'])
    else:
      found = self._xml.xpath(item['xpath'])

    len_found = len(found)

    if astype is bool:
      # handle the boolean flag case separately
      return bool(len_found)

    if not len_found: 
      # even for the case of numbers, do not set the value.  we 
      # want to detect "does not exist" vs. defaulting to 0
      # -- 2013-nov-19, JLS.
      return None   

    try:
      # added exception handler to catch malformed xpath expressesion
      # -- 2013-nov-19, JLS.
      # added support to handle multiple xpath values, i.e. a list of
      # things that have the same xpath expression (common in configs)
      # -- 2031-dec-06, JLS
      # added support to use the element tag if the text is empty
      def _munch(x):
        as_str = x if isinstance(x,str) else x.text
        if as_str is not None: as_str = as_str.strip()
        if not as_str: as_str = x.tag     # use 'not' to test for empty
        return astype(as_str)

      if 1 == len_found: return _munch( found[0] )
      return [_munch(this) for this in found ]

    except:
      raise RuntimeError("Unable to handle field:'%s'" % name)

    # and if we are here, then we didn't handle the field.
    raise RuntimeError("Unable to handle field:'%s'" % name)

  def __getitem__(self,name):
    """ 
    allow the caller to extract field values using :obj['name']:
    the same way they would do :obj.name:
    """
    return getattr(self,name)







# 3rd-party modules
from lxml.builder import E 

# module packages
from ... import Resource
from .... import jxml as JXML
from .nat_st_rule import NatStaticRule

class NatStaticRuleSet( Resource ):
  """
  [edit security nat static rule-set <name>]
  """

  PROPERTIES = [
    "description",
    "zone_from",
    "$rules",
    "$rules_count"
  ]

  def __init__(self, junos, name=None, **kvargs ):
    if name is None:
      # resource-manager
      Resource.__init__( self, junos, name, **kvargs )
      return

    self.rule = NatStaticRule( junos, M=self, parent=self )
    self._manages = ['rule']
    Resource.__init__( self, junos, name, **kvargs )

  ##### -----------------------------------------------------------------------
  ##### XML read
  ##### -----------------------------------------------------------------------

  def _xml_at_top(self):
    return E.security(E.nat(E.static(
      E('rule-set',E.name(self._name))
    )))

  def _xml_hook_read_begin(self,xml):
    """
    need to add the <from>,<to> elements to pick up the zone context
    need to add the rules, names-only
    """
    rs = xml.find('.//rule-set')
    rs.append(E('from'))
    rs.append(E.rule(JXML.NAMES_ONLY))
    return True

  def _xml_at_res(self, xml):
    return xml.find('.//rule-set')

  def _xml_to_py(self, as_xml, to_py ):
    Resource._r_has_xml_status( as_xml, to_py )
    Resource.copyifexists( as_xml, 'description', to_py)
    to_py['zone_from'] = as_xml.find('from/zone').text
    to_py['$rules'] = [rule.text for rule in as_xml.xpath('.//rule/name')]
    to_py['$rules_count'] = len(to_py['$rules'])

  ##### -----------------------------------------------------------------------
  ##### XML write
  ##### -----------------------------------------------------------------------

  def _xml_change_zone_from( self, xml ):
    xml.append(E('from', JXML.REPLACE, E.zone( self.should['zone_from'])))
    return True

  ##### -----------------------------------------------------------------------
  ##### Resource List, Catalog
  ##### -- only executed by 'manager' resources
  ##### -----------------------------------------------------------------------

  def _r_list(self):
    get = E.security(E.nat(E.static(
      E('rule-set', JXML.NAMES_ONLY)
    )))
    got = self.D.rpc.get_config( get )
    self._rlist = [name.text for name in got.xpath('.//name')]

  def _r_catalog(self):
    get = E.security(E.nat(E.static(
      E('rule-set')
    )))
    got = self.D.rpc.get_config( get )
    for ruleset in got.xpath('.//rule-set'):
      name = ruleset.find("name").text
      self._rcatalog[name] = {}
      self._xml_to_py( ruleset, self._rcatalog[name] )
    



# 3rd-party modules
from lxml.builder import E 

# module packages
from ... import Resource
from .... import jxml as JXML

class NatSrcRule( Resource ):
  """
  [edit security nat source rule-set <ruleset-name> rule <rule-name>]
  """

  PROPERTIES = [
   "match_src_addr",
   "match_dst_addr",
   "pool"
  ]

  ##### -----------------------------------------------------------------------
  ##### XML read
  ##### -----------------------------------------------------------------------

  def _xml_at_top(self):
    return E.security(E.nat(E.source(
      E('rule-set',
        E.name(self.P._name),
        E.rule(E.name(self._name))
      )
    )))

  def _xml_at_res(self, xml):
    return xml.find('.//rule')

  def _xml_to_py(self, as_xml, to_py ):
    """
    converts Junos XML to native Python
    """
    Resource._r_has_xml_status( as_xml, to_py )
    e = as_xml.find('src-nat-rule-match')
    to_py['match_src_addr'] = e.find('source-address').text
    to_py['match_dst_addr'] = e.find('destination-address').text
    to_py['pool'] = as_xml.find('.//pool-name').text

  ##### -----------------------------------------------------------------------
  ##### XML write
  ##### -----------------------------------------------------------------------

  def _xml_hook_build_change_begin( self, xml ):
    """
    when doing a write, assign default values if they are not present
    """
    def _default_to(prop, value):
      if prop not in self.should: self.should[prop] = value

    if self.is_new:
      _default_to( 'match_dst_addr', '0.0.0.0/0')
      _default_to( 'match_src_addr', '0.0.0.0/0')

  def _xml_change_match_src_addr(self, xml):
    xml.append(E('src-nat-rule-match',
      E('source-address', JXML.REPLACE, self.should['match_src_addr'])
    ))
    return True

  def _xml_change_match_dst_addr(self, xml):
    xml.append(E('src-nat-rule-match',
      E('destination-address', JXML.REPLACE, self.should['match_dst_addr'])
    ))
    return True

  def _xml_change_pool(self, xml):
    xml.append(E.then(
      E('source-nat',E.pool(E('pool-name', self.should['pool'])))
    ))
    return True

  ##### -----------------------------------------------------------------------
  ##### Manager List, Catalog
  ##### -----------------------------------------------------------------------

  def _r_list(self):
    self._rlist = self.P['$rules']

  def _r_catalog(self):
    get = E.security(E.nat(E.source(
      E('rule-set',
        E.name(self.P._name),
      )
    )))
    got = self.D.rpc.get_config( get )
    for rule in got.xpath('.//rule'):
      name = rule.find('name').text
      self._rcatalog[name] = {}
      self._xml_to_py( rule, self._rcatalog[name] )



# 3rd-party modules
from lxml.builder import E 

# module packages
from . import Resource
from .. import jxml as JXML
from .user_ssh_key import UserSSHKey 

class User( Resource ):
  """
  [edit system login user <name>]

  Resource name: str
    <name> is the user login name

  Manages resources:
    sshkey, UserSSHKey     
  """

  PROPERTIES = [
    'uid',  
    'fullname',       # the full-name field
    'userclass',      # user class
    'password',       # write-only clear-text password, will get crypt'd
    '$password',      # read-only crypt'd password
    '$sshkeys',       # read-only names of ssh-keys
  ]

  MANAGES = { 'sshkey' : UserSSHKey }

  ##### -----------------------------------------------------------------------
  ##### XML readers
  ##### -----------------------------------------------------------------------

  def _xml_at_top(self):
    return E.system(E.login(E.user(E.name( self._name ))))

  def _xml_at_res(self, xml):
    return xml.find('.//user')

  def _xml_to_py(self, has_xml, has_py ):
    Resource._r_has_xml_status( has_xml, has_py )

    has_py['userclass'] = has_xml.findtext('class')

    Resource.copyifexists( has_xml, 'full-name', has_py, 'fullname' )
    
    Resource.copyifexists( has_xml, 'uid', has_py )
    if 'uid' in has_py: has_py['uid'] = int(has_py['uid'])

    auth = has_xml.find('authentication')
    if auth is not None:
      # plain-text password
      Resource.copyifexists( auth, 'encrypted-password', has_py, '$password')

      # ssh-keys
      sshkeys = auth.xpath('ssh-rsa | ssh-dsa')
      if sshkeys is not None:
        has_py['$sshkeys'] = [(sshkey.tag, sshkey.findtext('name').strip())
          for sshkey in sshkeys
        ]

  ##### -----------------------------------------------------------------------
  ##### XML property writers
  ##### -----------------------------------------------------------------------

  def _xml_change_fullname(self, xml):
    xml.append(E('full-name', self['fullname']))
    return True

  def _xml_change_userclass(self, xml):
    xml.append(E('class', self['userclass']))
    return True

  def _xml_change_password(self,xml):
    xml.append(E.authentication(
      E('plain-text-password-value', self['password'])
    ))
    return True

  def _xml_change_uid(self,xml):
    xml.append(E.uid( str(self['uid'] )))
    return True

  ##### -----------------------------------------------------------------------
  ##### Manager List, Catalog
  ##### -----------------------------------------------------------------------

  def _r_list(self):
    get = E.system(E.login(E.user( JXML.NAMES_ONLY )))
    got = self.R.get_config( get )
    self._rlist = [name.text for name in got.xpath('.//user/name')]




import os
import sys

sys.path.insert(0,'lib')
from setuptools import setup, find_packages

requirements = [ 'netaddr', 'jinja2 >= 2.7.1', 'lxml >= 3.2.4',
                 'scp >= 0.7.0', 'PyYAML >= 3.10' ]

setup(
    name = "junos-eznc",
    namespace_packages = ['jnpr'],
    version = "0.0.4",
    author = "Jeremy Schulman",
    author_email = "jschulman@juniper.net",
    description = ( "Junos 'EZ' automation for non-programmers" ),
    license = "Apache 2.0",
    keywords = "Junos NETCONF networking automation",
    url = "http://www.github.com/Juniper/py-junos-eznc",
    package_dir={'':'lib'},    
    packages=find_packages('lib'),
    package_data={
        'jnpr.junos.op': ['*.yml'],
        'jnpr.junos.cfgro': ['*.yml']
    },
    install_requires=requirements,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: System Administrators',
        'Intended Audience :: Telecommunications Industry',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Other Scripting Engines',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: System :: Networking',
        'Topic :: System :: Networking :: Firewalls',
        'Topic :: Text Processing :: Markup :: XML'
    ],
)



# 3rd-party modules
from lxml.builder import E 

# module packages
from ... import Resource
from .... import jxml as JXML

class NatSrcPool( Resource ):
  """
  [edit security nat source pool <name>]
  """

  PROPERTIES = [
    'addr_from',
    'addr_to'
  ]

  def _xml_at_top(self):
    """
    configuration to retrieve resource
    """
    return E.security(E.nat(E.source(E.pool(E.name( self.name )))))

  ##### -----------------------------------------------------------------------
  ##### XML read
  ##### -----------------------------------------------------------------------

  def _xml_at_res(self, xml):
    """
    return Element at resource
    """
    return xml.find('.//pool')

  def _xml_to_py(self, as_xml, to_py ):
    """
    converts Junos XML to native Python
    """
    Resource._r_has_xml_status( as_xml, to_py )
    to_py['addr_from'] = as_xml.find('address/name').text
    to_py['addr_to'] = as_xml.find('address/to/ipaddr').text

  ##### -----------------------------------------------------------------------
  ##### XML property writers
  ##### -----------------------------------------------------------------------

  def _xml_change_addr_from(self, xml):
    # we need to always set the address/name given the structure of the
    # Junos configuration XML, derp.

    addr_from = self.should.get('addr_from') or self.has.get('addr_from')
    xml.append(E.address(JXML.REPLACE, E.name( addr_from )))
    return True

  def _xml_change_addr_to(self, xml):
    # we must always include the addr_from, so if we didn't expliclity
    # change it, we must do it now.

    if not self.should.has_key('addr_from'):
      self._xml_change_addr_from(xml)

    x_addr = xml.find('address')
    x_addr.append(E.to(E.ipaddr(self.should['addr_to'])))
    return True

  ##### -----------------------------------------------------------------------
  ##### Resource List, Catalog
  ##### -- only executed by 'manager' resources
  ##### -----------------------------------------------------------------------

  def _r_list(self):
    """
      build the policy context resource list from the command~
      > show security policies zone-context
    """
    get = E.security( E.nat( E.source(
      E.pool(JXML.NAMES_ONLY))))

    got = self.D.rpc.get_config( get )
    self._rlist = [name.text for name in got.xpath('.//pool/name')]

  def _r_catalog(self):
    get = E.security( E.nat( E.source( E.pool )))
    got = self.D.rpc.get_config( get )
    for pool in got.xpath('.//pool'):
      name = pool.find('name').text
      self._rcatalog[name] = {}
      self._xml_to_py( pool, self._rcatalog[name] )



from __future__ import absolute_import

import paramiko
from scp import SCPClient

class SCP(object):
  
  def __init__(self, junos, **scpargs):
    """
    constructor that wraps a paramiko 'scp' object.  
    """
    self._junos = junos
    self._scpargs = scpargs

  def open(self, **scpargs):
    """
    creates an instance of the scp object and return to caller for use
    """
    #@@@ should check for multi-calls to connect to ensure we don't keep opening
    #@@@ new connections
    junos = self._junos
    self._ssh = paramiko.SSHClient()
    self._ssh.load_system_host_keys()
    self._ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())

    # use junos._hostname since this will be correct if we are going
    # through a jumphost.

    self._ssh.connect( hostname = junos._hostname, 
      port=(22,int(junos._port))[junos._hostname=='localhost'],
      username=junos._auth_user,
      password=junos._auth_password,
    )    
    return SCPClient(self._ssh.get_transport(), **scpargs)

  def close(self):
    """
    closes the ssh/scp connection to the device
    """
    self._ssh.close()

  ### -------------------------------------------------------------------------
  ### CONTEXT MANAGER
  ### -------------------------------------------------------------------------

  def __enter__(self):
    return self.open(**self._scpargs)

  def __exit__(self,exc_ty,exc_val,exc_tb):
    self.close()    



# stdlib
import hashlib
import re
from os import path

# 3rd-party modules
from lxml.builder import E

# local modules
from .util import Util
from .scp import SCP 

__all__ = ['SW']

def _hashfile(afile, hasher,blocksize=65536):
  buf = afile.read(blocksize)
  while len(buf) > 0:
      hasher.update(buf)
      buf = afile.read(blocksize)
  return hasher.hexdigest()      

class SW(Util):
  """
  Softawre Utility class, used to perform a software upgrade and associated functions.

  Primary methods:
    install - perform the entire software installation process
    reboot - reboots the system for the new image to take effect    
    poweroff - shutdown the system

  Helpers called from install, but you can use these individually if needed:
    put - SCP put package file onto Junos device
    pkgadd - performs the 'request' operation to install the package
    validate - performs the 'request' to validate the package

  Other utils:
    rollback - same as 'request softare rollback'
    inventory - (property) provides file info for current and rollback 
                images on the device
  """

  def __init__(self, dev):
    Util.__init__(self, dev)
    self._RE_list = [x for x in dev.facts.keys() if x.startswith('version_RE')]
    self._multi_RE = bool(len(self._RE_list) > 1)
    self._multi_MX = bool(dev.facts['personality'] == "MX" and self._multi_RE is True)
    self._multi_VC = bool(self._multi_RE is True and dev.facts.get('vc_capable') is True)

  ##### -----------------------------------------------------------------------
  ##### CLASS METHODS
  ##### -----------------------------------------------------------------------

  @classmethod
  def local_sha256(cls,package):
    """
    computes the SHA-256 value on the package file.

    :package:
      complete path to the package (*.tgz) file on the local server      
    """
    return _hashfile(open(package,'rb'),hashlib.sha256())

  @classmethod
  def local_md5(cls,package):
    """
    computes the MD5 checksum value on the local package file.  

    :package:
      complete path to the package (*.tgz) file on the local server      
    """
    return _hashfile(open(package,'rb'),hashlib.md5())

  @classmethod
  def local_sha1(cls,package):
    """
    computes the SHA1 checksum value on the local package file.  

    :package:
      complete path to the package (*.tgz) file on the local server      
    """
    return _hashfile(open(package,'rb'),hashlib.sha1())    

  @classmethod
  def progress(cls,dev,report):
    """ simple progress report function """
    print dev.hostname + ": " + report

  ### -------------------------------------------------------------------------
  ### put - SCP put the image onto the device
  ### -------------------------------------------------------------------------

  def put(self, package, remote_path='/var/tmp', progress=None):
    """
    SCP 'put' the package file from the local server to the remote device.

    :package:
      file path to the package file on the local file system

    :remote_path:
      the directory on the device where the package will be copied to

    :progress:
      callback function to indicate progress.  You can use :SW.progress:
      for basic reporting.  See that class method for details.
    """
    def _progress(report): 
      # report progress only if a progress callback was provided
      if progress is not None: progress(self._dev, report)

    def _scp_progress(_path, _total, _xfrd):
      # init static variable
      if not hasattr(_scp_progress,'by10pct'): _scp_progress.by10pct = 0

      # calculate current percentage xferd
      pct = int(float(_xfrd)/float(_total) * 100)

      # if 10% more has been copied, then print a message
      if 0 == (pct % 10) and pct != _scp_progress.by10pct:
        _scp_progress.by10pct = pct
        _progress("%s: %s / %s (%s%%)" % (_path,_xfrd,_total,str(pct)))

    # execute the secure-copy with the Python SCP module

    with SCP(self._dev, progress=_scp_progress) as scp:
      scp.put(package, remote_path)

  ### -------------------------------------------------------------------------
  ### pkgadd - used to perform the 'request system software add ...'
  ### -------------------------------------------------------------------------

  def pkgadd(self, remote_package, **kvargs ):    
    """ 
    Issue the 'request system software add' command on the package.  No 
    validate is auto-set.  If you want to validate the image, do that 
    using the specific :validate(): method.  Also, if you want to
    reboot the device, suggest using the :reboot(): method rather
    than kvargs['reboot']=True.

    :remote_package:
      the file-path to the install package

    :kvargs:
      any additional parameters to the 'request' command can
      be passed within kvargs, following the RPC syntax
      methodology (dash-2-underscore,etc.)
    """

    args = dict(no_validate=True, package_name=remote_package)
    args.update(kvargs)

    dev_to = self.dev.timeout     # store device/rpc timeout
    self.dev.timeout = 60*60      # hardset to 1 hr for long running process
    rsp = self.rpc.request_package_add(**args)
    self.dev.timeout = dev_to     # restore original timeout

    got = rsp.getparent()
    rc = int(got.findtext('package-result').strip())
    return True if rc == 0 else got.findtext('output').strip()

  ### -------------------------------------------------------------------------
  ### validate - perform 'request' operation to validate the package
  ### -------------------------------------------------------------------------

  def validate(self, remote_package):
    """ issues the 'request' operation to validate the package against the config """
    rsp = self.rpc.request_package_validate(package_name=remote_package).getparent()
    errcode = int(rsp.findtext('package-result'))
    return True if 0 == errcode else rsp.findtext('output').strip()

  def remote_checksum(self, remote_package):
    """ computes the MD5 checksum on the remote device """
    rsp = self.rpc.get_checksum_information(path=remote_package)
    return rsp.findtext('.//checksum').strip()    

  ### -------------------------------------------------------------------------
  ### safe_copy - copies the package and performs checksum
  ### -------------------------------------------------------------------------

  def safe_copy(self, package, **kvargs):
    """
    Copy the install package safely to the remote device.  By default
    this means to clean the filesystem to make space, perform the 
    secure-copy, and then verify the MD5 checksum.

    For :kvargs: values, please refer to the :install(): method.
    """
    remote_path = kvargs.get('remote_path','/var/tmp')
    progress = kvargs.get('progress')
    checksum = kvargs.get('checksum')
    cleanfs = kvargs.get('cleanfs', True)

    def _progress(report): 
      if progress is not None: progress(self._dev, report)

    if checksum is None: 
      _progress('computing local checksum on: %s' % package)
      checksum = SW.local_md5(package)

    if cleanfs is True:
      dto = self.dev.timeout
      self.dev.timeout = 5 * 60
      _progress('cleaning filesystem ...')
      self.rpc.request_system_storage_cleanup()
      self.dev.timeout = dto

    # we want to give the caller an override so we don't always
    # need to copy the file, but the default is to do this, yo!
    self.put( package, remote_path, progress)

    # validate checksum:
    remote_package = remote_path + '/' + path.basename(package)
    _progress('computing remote checksum on: %s' % remote_package)
    remote_checksum = self.remote_checksum(remote_package)

    if remote_checksum != checksum:
      _progress("checksum check failed.")
      return False
    _progress("checksum check passed.")    

    return True

  ### -------------------------------------------------------------------------
  ### install - complete installation process, but not reboot
  ### -------------------------------------------------------------------------

  def install(self, package, remote_path='/var/tmp', progress=None,
    validate=False, checksum=None, cleanfs=True, no_copy=False):
    """
    Performs the complete installation of the :package: that includes the following steps:
    When no_copy is False (default-case)
      (1) computes the local MD5 checksum if not provided in :checksum:
      (2) performs a storage cleanup if :cleanfs: is True
      (3) SCP copies the package to the :remote_path: directory
      (4) computes the remote MD5 checksum and matches it to the local value
    Then:
      (5) validates the package if :validate: is True
      (6) installs the package

    You can get a progress report on this process by providing a :progress: callback;
    see description below.

    You will need to invoke the :reboot(): method explicity to reboot the device.

    :package: 
      is the install package tarball on the local filesystem.
    
    :remote_path: 
      is the directory on the Junos device where the package file will be SCP'd to.  
    
    :validate:
      determines whether or not to perform a config validation against the new image

    :checksum:
      MD5 hexdigest of the package file.  If this is not provided, then this
      method will perform the calculation.  If you are planning on using the
      same image for multiple updates, you should consider using the :local_md5():
      method to precalculate this value and then provide to this method.

    :cleanfs:
      determines whether or not to perform a 'storeage cleanup' before SCP'ing 
      the file to the device.

    :progress:
      if provided, this is a callback function with a function prototype given
      the Device instance and the report string, e.g.
        
        def myprogress(dev, report):
          print "host: %s, report: %s" % (dev.hostname, report)
    """
    def _progress(report): 
      if progress is not None: progress(self._dev, report)
    
    rpc = self.rpc

    ### -----------------------------------------------------------------------
    ### perform a 'safe-copy' of the image to the remote device
    ### -----------------------------------------------------------------------

    if no_copy is False:
      copy_ok = self.safe_copy(package, remote_path=remote_path, progress=progress,
        cleanfs=cleanfs, checksum=checksum)
      if copy_ok is False: return False

    ### -----------------------------------------------------------------------
    ### at this point, the file exists on the remote device
    ### -----------------------------------------------------------------------

    remote_package = remote_path + '/' + path.basename(package)

    if validate is True:
      _progress("validating software against current config, please be patient ...")
      v_ok = self.validate(remote_package)
      if v_ok is not True:
        return v_ok # will be the string of output

    if self._multi_RE is False:
      # simple case of device with only one RE
      _progress("installing software ... please be patient ...")
      return self.pkgadd( remote_package )
    else:
      # we need to update multiple devices
      if self._multi_MX is True:
        ok = True
        _progress("installing software on RE0 ... please be patient ...")
        ok &= self.pkgadd( remote_package, re0=True )
        _progress("installing software on RE1 ... please be patient ...")
        ok &= self.pkgadd( remote_package, re1=True )
        return ok
      elif self._multi_VC is True:
        ok = True
        # extract the VC number out of the 'version_RE<n>' string
        vc_members = [re.search('(\d+)',x).group(1) for x in self._RE_list]
        for vc_id in vc_members:
          _progress("installing software on VC member: {} ... please be patient ...".format(vc_id))
          ok &= self.pkgadd( remote_package, member=vc_id)
        return ok

  ### -------------------------------------------------------------------------
  ### rebbot - system reboot
  ### -------------------------------------------------------------------------

  def reboot(self, in_min=0):    
    """ 
    Perform a system reboot, with optional delay (in minutes).  

    If the device is an MX with dual-RE installed, then both RE will be
    rebooted.
    """
    cmd = E('request-reboot', E('in', str(in_min)))

    if self._multi_MX is True:
      cmd.append(E('both-routing-engines'))

    rsp = self.rpc(cmd)
    got = rsp.getparent().findtext('.//request-reboot-status').strip()
    return got

  ### -------------------------------------------------------------------------
  ### poweroff - system shutdown
  ### -------------------------------------------------------------------------

  def poweroff(self, in_min=0):    
    """
    Perform a system shutdown, with optional delay (in minutes) .

    If the device is an MX with dual-RE installed, then both RE will be
    rebooted.    
    """
    cmd = E('request-power-off', E('in', str(in_min)))

    if self._multi_MX is True:
      cmd.append(E('both-routing-engines'))

    rsp = self.rpc(cmd)
    got = rsp.getparent().findtext('.//request-reboot-status').strip()
    return got

  ### -------------------------------------------------------------------------
  ### rollback - clears the install request
  ### -------------------------------------------------------------------------

  def rollback(self):
    """ 
    issues the 'request' command to do the rollback and returns the string
    output of the results
    """
    rsp = self.rpc.request_package_rollback()
    return rsp.text.strip()

  ### -------------------------------------------------------------------------
  ### inventory - file info on current and rollback packages
  ### -------------------------------------------------------------------------

  @property
  def inventory(self):
    """
    returns a dictionary of file listing information for current and rollback
    Junos install packages.  This information comes from the /packages directory.
    """
    from .fs import FS 
    fs = FS(self.dev)
    pkgs = fs.ls('/packages') 
    return dict(current=pkgs['files'].get('junos'), rollback=pkgs['files'].get('junos.old'))



from .nat_src_pool import NatSrcPool
from .nat_src_ruleset import NatSrcRuleSet
from .nat_src_rule import NatSrcRule

from .nat_st_ruleset import NatStaticRuleSet
from .nat_st_rule import NatStaticRule
from .nat_proxy_arp import NatProxyArp


from lxml.builder import E
from ..import jxml as JXML

def chassis(junos, facts):
  """ 
  Obtain basic chassis facts:
    model : product model
    serialnumber : serial number 
    hostname : host 
    fqdn : fully-qualified domain-name
    domain : domain-name

  NOTES:
  (1) if in a 2RE system, this routine will only load the information
      from the first chassis item.
  (2) hostname, domain, and fqdn are retrieved from configuration data;
      inherited configs are checked.
  """
  rsp = junos.rpc.get_chassis_inventory()

  if rsp.tag == 'multi-routing-engine-results':
    facts['2RE'] = True
    x_ch = rsp.xpath('.//chassis-inventory')[0].find('chassis')
  else:
    facts['2RE'] = False
    x_ch = rsp.find('chassis')

  facts['model'] = x_ch.find('description').text
  try:
    facts['serialnumber'] = x_ch.find('serial-number').text
  except:
    # if the toplevel chassis does not have a serial-number, then
    # check the Backplane chassis-module
    facts['serialnumber'] = x_ch.xpath('chassis-module[name="Backplane"]/serial-number')[0].text

  got = junos.rpc.get_config(
    E.system(
      E('host-name'),
      E('domain-name')
    ),
    JXML.INHERIT
  )

  hostname = got.find('.//host-name')
  facts['hostname'] = hostname.text if hostname is not None else 'Amnesiac'
  facts['fqdn'] = facts['hostname']

  domain = got.find('.//domain-name')
  if domain is not None: 
    facts['domain'] = domain.text
    facts['fqdn'] += '.%s' % facts['domain']    
  else:
    facts['domain'] = None

